/*
 * mfe_nn.h
 *
 *  Created on: Mar 12, 2015
 *      Author: christian
 */

#ifndef MFE_NN_H_
#define MFE_NN_H_


using namespace std;

#include <vector>
#include <cmath>

#include "SystemClass.h"
#include "simplified_basic_equations.h"


void System::calcNewLambda()
{
	vector<long double> phases;
	long double gammaA,gammaB,invW,h,sumInvW;
	int N = this->N_1*this->N_2;

	long double minLambda=0.0L;

	for(int n_1=0; n_1<this->N_1; n_1++)
	{
		for(int n_2=0; n_2<this->N_2; n_2++)
		{

			phases=this->phaseMatrix[n_1][n_2];
			gammaA=this->A*calcGA(phases);
			gammaB=this->B*calcGB(phases);

			if(minLambda<abs(gammaA)-gammaB)
			{
				minLambda=abs(gammaA)-gammaB;
			}
		}
	}

	long double gap=1.0L;
	long double oldGap=1.0L;
	long double l=minLambda+0.00000000000000001L; //To avoid nan in further calc
	long double delta=1.0L;

	int iter=0;

	while(abs(gap)>0.0000000000001L && iter<500)	//0.000001
	{
		h=0.0L;
		sumInvW=0.0L;
		for(int n_1=0; n_1<this->N_1; n_1++)
		{
			for(int n_2=0; n_2<this->N_2; n_2++)
			{
				phases=this->phaseMatrix[n_1][n_2]; //[k*d-, k*d_0, k*d+].

				gammaA=this->A*calcGA(phases);	//Calculate all necessary quantities using the current (old) generation
				gammaB=this->B*calcGB(phases);
				invW=calcInvW(gammaA,gammaB,l);


				sumInvW+=invW;

				h-=gammaB*invW;

			}
		}

		h+=2*N*(this->S+0.5L);
		h/=sumInvW;

		gap=l-h;


		if(gap*oldGap<0)	//If we have crossed the intersection, delta is set to be smaller and changes sign (aka we go backwards)
		{
			delta*=-0.1L;
		}

		l+=delta;
		oldGap=gap;

		if(l<minLambda)
		{
			cout<<"No exact solution, setting lambda to minLambda..."<<endl;
			cout<<gap<<endl;
			cout<<iter<<endl;
			this->lambda=minLambda+0.00000000000000001L;
			return;
		}

		iter++;
	}

	if(iter==500)
		cout<<"Gap too small! Using current solution (500 iterations) with gap: "<<gap<<endl;

	this->lambda=l;

	return;
}

void System::solveEquations(int const &numbOfGen)	//Numerical solution of the MF-equations. numbOfGen is the number of iterations to be performed.
{

	long double invN = (long double) 1/(this->N_1*this->N_2);
	long double gammaA, gammaB, invW;
	vector<long double> phases;	//will hold values of k*d

	long double newA=0.0L;	//Next generation of A, B and Lambda
	long double newB=0.0L;

	this->calcNewLambda();

	for(int gen=0; gen<numbOfGen; gen++)
	{
		for(int n_1=0; n_1<this->N_1; n_1++)	//Looping through k
		{
			for(int n_2=0; n_2<this->N_2; n_2++)
			{
				phases=this->phaseMatrix[n_1][n_2]; //[k*d-, k*d_0, k*d+].

				gammaA=this->A*calcGA(phases);	//Calculate all necessary quantities using the current (old) generation
				gammaB=this->B*calcGB(phases);
				invW=calcInvW(gammaA,gammaB,this->lambda);

				newA+=gammaA*gammaA*invW;
				newB+=(gammaB+this->lambda)*calcGB(phases)*invW;


			}
		}

		newA*=(invN/(6*this->A));
		newB*=(invN/6);

		this->A=newA;
		this->B=newB;
		this->calcNewLambda();

		newA=0.0L;
		newB=0.0L;

		//cout<<setprecision(13)<<gen<<"  "<<"A: "<<this->A<<", B: "<<this->B<<", lambda:"<<this->lambda<<endl<<endl;

	}


	return;
}

long double System::calcE()
{
	long double gammaA, gammaB;
	vector<long double> phases;
	int N = this->N_1*this->N_2;

	long double E=0.0L;
	for(int n_1=0; n_1<this->N_1; n_1++)
	{
		for(int n_2=0; n_2<this->N_2; n_2++)
		{
			phases=this->phaseMatrix[n_1][n_2];
			gammaA=this->A*calcGA(phases);
			gammaB=this->B*calcGB(phases);

			E+=calcW(gammaA,gammaB,this->lambda);
		}
	}

	E*=0.5L;
	E+=-1.5L*this->S*N-N*this->lambda*(this->S+0.5L);


	return(E/N);
}







#endif /* MFE_NN_H_ */
